(*  Title:      AuxiliaryFunctions.thy
    Author:     Ludovic Henrio and Florian Kammuller
                2014
    Note:       Functions on Maps for Multi-active object formalisation
                Plus generic lemmas (not related to MAO)
*)

header {* Auxiliary functions for Multi-ASP *}

theory AuxiliaryFunctions imports Main begin


definition
  remove_from_map :: "('a ~=> 'b) => 'a  => ('a ~=> 'b)"  (infixl "|``"  110) where
  "m|``x = m|`(dom m - {x})"

definition listunionMap:: "('a \<Rightarrow> 'b set) \<Rightarrow> ('a list) \<Rightarrow> 'b set"
where
  "listunionMap f l \<equiv> (foldr (op  \<union>) (map f l) {})"

lemma listunionMap_Empty[simp]: "listunionMap f [] = {}"
by (auto simp: listunionMap_def)

(*
lemma listunionMap_Empty_fun[simp]: "listunionMap (\<lambda>x.{}) (L::'a list) = {}"
apply (induct_tac L)
apply (auto simp: listunionMap_def)
done
*)

lemma listunionMap_Empty_fun[simp]: "listunionMap (\<lambda>x.{}) (L::'a list) = {}"
by (induct L, auto simp: listunionMap_def)
lemma listunionMap_Single[simp]: "listunionMap f [x] = f x"
by ( auto simp: listunionMap_def)

lemma listunionMap_Cons[simp]: "listunionMap f (x#L) = (f x) \<union> (listunionMap f L)"
by ( auto simp: listunionMap_def)

lemma map_of_zip_empty[rule_format,simp]: "ran (map_of (zip [] Y)) = {}"
by auto
lemma map_of_zip_replicate[rule_format,intro]: "x\<in>set vl \<longrightarrow>map_of (zip vl (replicate (length vl) Y)) x = Some Y"
by (induct_tac vl, auto)

lemma map_of_zip_outside_dom[rule_format,intro]: "x\<notin>set vl \<longrightarrow>map_of (zip vl (replicate (length vl) Y)) x = None"
by (induct_tac vl, auto)

lemma map_of_zip_replicate_simp[rule_format,simp]: 
    "map_of (zip vl (replicate (length vl) Y)) x = (if (x\<in>set vl) then Some Y else None)"
by (case_tac "(x\<in>set vl)",  auto simp: map_of_zip_replicate map_of_zip_outside_dom)

lemma map_of_zip_replicate_map_simp[rule_format,simp]: 
    "map_of (zip (map f vl) (replicate (length vl) Y)) x = (if (x\<in>set (map f vl)) then Some Y else None)"
apply (subgoal_tac "length vl = length (map f vl)")
apply (erule ssubst)
apply (simp only: map_of_zip_replicate_simp)
apply simp
done

lemma ran_map_of_zip_repl: "ran (map_of (zip (v#vl) (replicate (length (v#vl)) Y))) = {Y}"
by (auto simp: ran_def)
lemma ran_map_of_zip_2: "N=length L\<Longrightarrow>if (N=0) then ran (map_of (zip L (replicate N Y))) = {} else ran (map_of (zip L (replicate N Y))) = {Y}"
by (cases, auto simp: ran_def)


lemma ran_update: "ran (M(x\<mapsto>V)) \<subseteq> ran M \<union> {V}"
by (auto simp: ran_def)

lemma map_of_zip: "length xs=length ys \<Longrightarrow>(x \<in> set xs) = (\<exists>y\<in>set ys. map_of (zip xs ys) x = Some y)"
apply (induct rule: list_induct2)
apply (auto)
apply (drule Map.map_of_zip_is_None,auto)
done

lemma ran_map_of_zip: "length xs=length ys \<Longrightarrow> map_of (zip xs ys) x = Some y \<Longrightarrow>y\<in>set ys"
by (frule map_of_zip_is_Some, drule map_of_zip, auto)

primrec allTrue_list
where
  "allTrue_list Nil = True" |
  "allTrue_list (a#L) = (a\<and>allTrue_list L)"

lemma allTrue_list_nth: "allTrue_list L =(\<forall>n<length L. L!n)"
apply (induct_tac L)
apply auto
apply (case_tac n, auto)
done
(* abbreviation allTrue_list
where
"allTrue_list L \<equiv>\<forall>n<length L. L!n" *)

lemma allTrue_list_set: "allTrue_list L \<Longrightarrow>a\<in>set L \<Longrightarrow> a"
by (induct L,auto)
lemma allTrue_list_set_map: "allTrue_list (map f L) \<Longrightarrow>a\<in>set L \<Longrightarrow> f a"
by (induct L,auto)

lemma foldr_Un_larger: "x\<in> foldr (op \<union>) L (L') \<Longrightarrow> x \<in> foldr (op \<union>) L (a \<union> L')"
by (induct L, auto)
lemma foldr_Un_mapI[intro]: " {N} \<in> f ` set L \<Longrightarrow> N \<in> foldr op \<union> (map f L) S"
by (induct L, auto)

lemma foldr_Un_other: "x\<in> a \<Longrightarrow> x \<in> foldr (op \<union>) L (a \<union> L')"
by (induct L, auto)

lemma foldr_Un_mapD :
"x \<in> foldr op \<union> (map f L) {} \<Longrightarrow> \<exists>y\<in>set L. x\<in>f y" 
by (induct L, auto)
end
